#include<stdarg.h>
#include<stdlib.h>
#include<stddef.h>
#include<stdio.h>
#include<string.h>
#include<math.h>

#define HASH 256
#define MAXIMGS 1000
#define VERBOSE 0

int *dis,*fij;
int **mdis;


typedef struct pixel {
  int r;
  int g;
  int b;
} tpixel;

typedef struct patch {
  int sr;
  int sg;
  int sb;
  int c;
  int used;
  void *sig;
} tpatch;

typedef struct nodo {
  int img;
  int d;
  void *sig;
} tnodo;

typedef struct rank {
  int d;
  int ind;
} trank;

tpixel **img;
tpatch ***vec;

int rows,cols,cimg;
tnodo **G;
int inf=100000000;


void memtest(void *ptr)
{
  if (ptr==NULL)
    {
      fprintf(stderr,"insufficient memory\n");
      exit(EXIT_FAILURE);
    }
}


void leeimagen(FILE *fi,int q) 
{
  int i,j;
  int maxval;
  char cad[100];


  fscanf(fi,"%s",cad);
  if (strcmp(cad,"P3")) {
    fprintf(stderr,"formato de imagen incorrecto\n");
    exit(1);
  }

  fscanf(fi,"%d %d",&cols,&rows);
  
  if (VERBOSE) fprintf(stderr,"%dx%d pixels\n",rows,cols);

  fscanf(fi,"%d",&maxval);
  if (maxval!=255) exit(1);


  img=(tpixel **)malloc(rows*sizeof(tpixel *));
  for(i=0;i<rows;i++)
    img[i]=(tpixel *)malloc(cols*sizeof(tpixel));
  
  for(i=0;i<rows;i++)
    for(j=0;j<cols;j++) {
      fscanf(fi,"%d %d %d ",&(img[i][j].r),&(img[i][j].g),&(img[i][j].b));
      img[i][j].r/=q;
      img[i][j].g/=q;
      img[i][j].b/=q;
      }
}

int busca(int sr,int sg,int sb,int img,int s,int orig)
{
  int h;
  tpatch *act;
  
  h=(sr+sg+sb)%HASH;

  act=vec[img][h];
  while (act!=NULL) {
    if ((act->sr==sr)&&(act->sg==sg)&&(act->sb==sb)) {
      act->used=orig;
      if (s) act->c++;
      return act->c;
    }
    act=act->sig;
  }
  return 0;

}


void addpatch(int sr,int sg,int sb) 
{
  tpatch *act,*new;
  int enc=0;
  int h;
  
  h=(sr+sg+sb)%HASH;

  if (!busca(sr,sg,sb,cimg,1,-1)) {
    new=(tpatch *)malloc(1*sizeof(tpatch));
    new->sr=sr;
    new->sg=sg;
    new->sb=sb;
    new->c=1;
    new->used=-1;

    new->sig=vec[cimg][h];
    vec[cimg][h]=new;

  }
  
 

}

void img2patch(int w) 
{
  int i,j,n,m,w2;
  int sr,sg,sb;

  w2=w/2;

  for(i=w2;i<rows-w2;i++)
    for(j=w2;j<cols-w2;j++) {
      sr=sg=sb=0;
      for(n=i-w2;n<=i+w2;n++)      
	for(m=j-w2;m<=j+w2;m++) {
	  sr+=img[n][m].r;
	  sg+=img[n][m].g;
	  sb+=img[n][m].b;
	}
      addpatch(sr,sg,sb);
    }
  

  for(i=0;i<rows;i++)
    free(img[i]);
  free(img);

}


int distancia(int a,int b) 
{
  int dab,dba,cab,c;
  tpatch *act;
  int h;
  dab=dba=cab=0;

  // cab, dab;
  for(h=0;h<HASH;h++) {
    act=vec[a][h];
    while (act!=NULL) {
      c=busca(act->sr,act->sg,act->sb,b,0,a);
      if (c==0) dab+=act->c;
      else cab+=abs(act->c-c);
      act=act->sig;
    }
  }

  // dba;
  for(h=0;h<HASH;h++) {
    act=vec[b][h];
    while (act!=NULL) {
      c=busca(act->sr,act->sg,act->sb,a,0,b);
      if (c==0) dba+=act->c;
      act=act->sig;
    }
  }
  

  return dab+dba+cab;

}


int compare(const void *_a, const void *_b) {
 
        trank *a, *b;
        
        a = (trank *) _a;
        b = (trank *) _b;
        
	if (b[0].d < a[0].d) return 1;
	else if (b[0].d > a[0].d) return -1;
 	// En caso de empate mirar los ndices.
	else if (a[0].ind < b[0].ind) return -1; 
	else return 1;
}

void addedge(int o,int d,int dist)
{
  tnodo *new;

  new=(tnodo *)malloc(1*sizeof(tnodo));
  new->img=d;
  new->d=dist;
  new->sig=G[o];

  G[o]=new;

}


void  Dijkstra(int s,int n) {

  int i,j,min,f;
  tnodo *act;
  
  
  for(i=0;i<n;i++) {
    dis[i]=inf;
    fij[i]=0;
  }
  
  
  dis[s]=0;
  for(i=0;i<n;i++) {

    min=inf; 
    for (j=0;j<n;j++)
      if ((!fij[j]) && (dis[j]<min)) { 
	min=dis[j];
	f=j;
      }
 
    if (min!=inf) {
      fij[f]=1;
      act=G[f];
      while(act!=NULL) {
	if ((!fij[act->img])&&(dis[act->img]>dis[f]+act->d)) dis[act->img]=dis[f]+act->d;
	act=act->sig;
      }
    }
  }

  for(i=0;i<n;i++) {
    if (dis[i]!=inf) 
      printf("%d ",dis[i]);
    else 
      printf("inf ");
  }
  printf("\n");

}





int main(int argc,char *argv[],char *env[])
{
  FILE *fe,*fi; 
  int k,w,q,i,j;
  char fname[2000];
  trank *dist;



  if (argc<5) {
    fprintf(stderr,"concurso file k w q\n");
    exit(1);
  }

  k=atoi(argv[2]);
  w=atoi(argv[3]);
  q=atoi(argv[4]);
 
  fe=fopen(argv[1],"rt");
  if (fe==NULL) {
    fprintf(stderr,"no se puede abrir fichero %s\n",argv[1]);
    exit(1);
  }

  vec=(tpatch ***)malloc(MAXIMGS*sizeof(tpatch *));
  for(i=0;i<MAXIMGS;i++) {
    vec[i]=(tpatch **)malloc(HASH*sizeof(tpatch*));
    for(j=0;j<HASH;j++)
      vec[i][j]=NULL;
  }

  cimg=0;
  while(!feof(fe)) {
    fscanf(fe,"%s",&(fname[0]));
    if (!feof(fe)) {
      if (VERBOSE) fprintf(stderr,"Abriendo %s ...\n",fname);
 
      fi=fopen(fname,"rt");
      if (fi==NULL) {
	fprintf(stderr,"no se puede abrir fichero %s\n",fname);
	exit(1);
      }
   
      if (VERBOSE) fprintf(stderr,"Leyendo %s ...\n",fname);
      leeimagen(fi,q);
      img2patch(w);

      fclose(fi);

      cimg++;
    }
  }

  fprintf(stderr,"%d imagenes leidas\n",cimg);
  


  // Crear grafo y vector distancias
  G=(tnodo **)malloc(cimg*sizeof(tnodo *));
  for(i=0;i<cimg;i++)
    G[i]=NULL;

  dist=(trank *)malloc(cimg*sizeof(trank));
  
  
  // Crear matriz de distancias (s que cabe :) )
  mdis=(int **)malloc(cimg*sizeof(int *));
  for(i=0;i<cimg;i++)
    mdis[i]=(int *)malloc(cimg*sizeof(int));
  
  fprintf(stderr,"Distancias:\n");

  for(i=0;i<cimg;i++) 
    for(j=i;j<cimg;j++) 
      if (i!=j) mdis[i][j]=mdis[j][i]=distancia(i,j);
      else mdis[i][i]=0;

  fprintf(stderr,"done\n");

  fe=fopen("dist.txt","wt");
  for(i=0;i<cimg;i++) {
    for(j=0;j<cimg;j++) {
      dist[j].d=mdis[i][j];
      dist[j].ind=j;
      fprintf(fe,"%d ",mdis[i][j]);
    }
    fprintf(fe,"\n");

    
    qsort(dist, cimg, sizeof(trank), &compare);
      
    for(j=1;j<=k;j++) {
      addedge(i,dist[j].ind,dist[j].d);
      addedge(dist[j].ind,i,dist[j].d);
    }
    
  }
  fclose(fe);
  fprintf(stderr,"\n");
 
  // Reserva para Dijkstra
  dis=(int *)malloc(cimg*sizeof(int));
  fij=(int *)malloc(cimg*sizeof(int));

  // Distancias dentro del grafo
  fprintf(stderr,"Dijkstra\n");
  for(i=0;i<cimg;i++) 
    Dijkstra(i,cimg);

  exit(1);
}
